/****************************************************************************
   MIDI Controller - A program that runs MIDI controller GUIs built in Glade
   
   Copyright (C) 2004  Lars Luthman <larsl@users.sourceforge.net>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#include <cerrno>
#include <cstdio>
#include <fstream>
#include <iostream>
#include <map>
#include <queue>
#include <list>
#include <sstream>
#include <string>

#include <gtkmm.h>
#include <sigc++/slot.h>
#include <libglademm.h>
#include <asoundlib.h>


using namespace std;
using namespace Gtk;
using namespace Gtk::Menu_Helpers;
using namespace Gnome::Glade;


class ControllerBox {
public:
  ControllerBox(string uiFile = "") throw(XmlError);

  /** Go through the widget tree and create a map of all Adjustment objects
      that should be used as MIDI controls. */
  map<string, Adjustment*> createAdjustmentMap(Widget* topWidget);
  
  /** Connect all MIDI control widgets to the MIDI CC sender. */
  void connectWidgets(map<string, Adjustment*>& widgets);

  /** Load the GUI, throw XmlError if it doesn't work. */
  Glib::RefPtr<Xml> loadGUI(string uiFile);

  /** Initialize the ALSA sequencer object, return false if it doesn't work. */
  bool initALSA();
  
  /** This function returns a pointer to the main window. */
  Window* getWindow() const;
  
  /** This function returns a pointer to an Adjustment, if the widget has
      one. */
  Adjustment* getAdjustment(Widget* widget) const;
  
  bool popupMenu(GdkEventKey* e);
  
  static void showAboutDialog();
  
  void savePreset() const;
  
  void loadPreset();
  
  void sendAll() const;
  
private:
  
  /** This class wraps the info about a single controller and can send
      MIDI CC events. */
  class Controller {
  public:
    /** This constructor does NOT create a valid object! Don't use it! */
    Controller();
    /** Use this one instead. */
    Controller(Adjustment* valueAdj, snd_seq_t* sequencer);
    
    void setChannel(unsigned char channel);
    void setChannelAdj(Adjustment* channelAdj);
    void setParameter(unsigned int parameter);
    void setParameterAdj(Adjustment* parameterAdj);
    void sendControlChange() const;
    void block();
    void unblock();
    
  private:
    
    void onChannelChange();
    void onParamChange();
    
    Adjustment* valA;
    Adjustment* chanA;
    Adjustment* paramA;
    unsigned char chn;
    unsigned int par;
    snd_seq_t* seq;
    sigc::connection conn;
  };
  
  snd_seq_t *seq;
  int seqPort;
  Window* win;
  map<string, Adjustment*> adjustments;
  vector<Controller*> controllers;
};



// CONTROLLERBOX IMPLEMENTATION

ControllerBox::ControllerBox(string uiFile) throw(XmlError) {
  Glib::RefPtr<Xml> refXml = loadGUI(uiFile);
  win = dynamic_cast<Window*>(refXml->get_widget("winMain"));
  if (!win) 
    throw XmlError("Could not find the main window widget!");
  adjustments = createAdjustmentMap(win);
  initALSA();
  connectWidgets(adjustments);
  win->add_events(Gdk::KEY_PRESS_MASK);
  win->signal_key_press_event().
    connect(sigc::mem_fun(this, &ControllerBox::popupMenu));
}


map<string, Adjustment*> ControllerBox::createAdjustmentMap(Widget* topWidget){
  map<string, Adjustment*> m;
  queue<Widget*> q;
  q.push(topWidget);
  while (!q.empty()) {
    Widget* w = q.front();
    q.pop();
    Adjustment* adj;
    if (w->get_name().find("mcw") == 0 && (adj = getAdjustment(w)))
      m[w->get_name()] = adj;
    Container* c = dynamic_cast<Container*>(w);
    if (c) {
      list<Widget*> l = (list<Widget*>)(c->get_children());
      while (l.size()) {
	q.push(*l.begin());
	l.pop_front();
      }
    }
  }
  return m;
}


void ControllerBox::connectWidgets(map<string, Adjustment*>& adjustments) {
  
  // templates for channel and parameter number widgets
  char chanName[] = "mcwC0";
  char paramName[] = "mcwP0";
  
  // iterate over all mcw widgets
  for (map<string, Adjustment*>::const_iterator iter = adjustments.begin();
       iter != adjustments.end(); ++iter) {
    int chan, param;
    char c1, c2;
    const char* name = iter->first.c_str();
    if (sscanf(name + 3, "%c%d%c%d", &c1, &chan, &c2, &param) == 4) {
      Adjustment* valueAdj = iter->second;
      Adjustment* chanAdj = NULL;
      Adjustment* paramAdj = NULL;
      if (c1 == 'c') {
	chanName[4] = '0' + chan;
	chanAdj = adjustments[chanName];
      }
      if (c2 == 'p') {
	paramName[4] = '0' + param;
	paramAdj = adjustments[paramName];
      }
      if ((c1 == 'C' || chanAdj != NULL) && (c2 == 'P' || paramAdj != NULL)) {
	Controller* cont = new Controller(valueAdj, seq);
	controllers.push_back(cont);
	if (chanAdj)
	  cont->setChannelAdj(chanAdj);
	else
	  cont->setChannel(chan);
	if (paramAdj)
	  cont->setParameterAdj(paramAdj);
	else
	  cont->setParameter(param);
	cerr<<"Connected "<<iter->first<<endl;
      }
      else {
	cerr<<"Could not connect "<<iter->first<<endl;
	continue;
      }
    }
  }
}


Glib::RefPtr<Xml> ControllerBox::loadGUI(string uiFile) {
  if (uiFile != "")
    return Xml::create(uiFile);
  else {
    try { return Xml::create("controller.glade"); }
    catch (XmlError& e) {
      cerr<<e.what()<<endl;
      return Xml::create(PREFIX "/share/midicontroller/controller.glade");
    }
  }
}


bool ControllerBox::initALSA() {
  if (snd_seq_open(&seq, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0) {
    cerr<<"Could not open ALSA sequencer: "<<snd_strerror(errno)<<endl;
    return false;
  }
  snd_seq_set_client_name(seq, "Controller");
  seqPort = snd_seq_create_simple_port(seq, "MIDI OUT",
				       SND_SEQ_PORT_CAP_READ | 
				       SND_SEQ_PORT_CAP_SUBS_READ,
				       SND_SEQ_PORT_TYPE_APPLICATION);
  if (seqPort < 0) {
    cerr<<"Error creating sequencer port: "<<snd_strerror(errno)<<endl;
    return false;
  }
  
  return true;
}


Adjustment* ControllerBox::getAdjustment(Widget* widget) const {
  Range* range = dynamic_cast<Range*>(widget);
  if (range)
    return range->get_adjustment();
  SpinButton* spin = dynamic_cast<SpinButton*>(widget);
  if (widget)
    return spin->get_adjustment();
  return NULL;
}


bool ControllerBox::popupMenu(GdkEventKey* e) {
  if (e->keyval == GDK_F1) {
    static Menu* m = NULL;
    if (m == NULL) {
      m = new Menu;
      MenuList ml = m->items();
      ml.push_back(MenuElem("Load preset", 
			    sigc::mem_fun(this, &ControllerBox::loadPreset)));
      ml.push_back(MenuElem("Save preset",
			    sigc::mem_fun(this, &ControllerBox::savePreset)));
      ml.push_back(SeparatorElem());
      ml.push_back(MenuElem("Send all CCs",
			    sigc::mem_fun(this, &ControllerBox::sendAll)));
      ml.push_back(SeparatorElem());
      ml.push_back(MenuElem("About MIDI Controller",
			    &ControllerBox::showAboutDialog));
    }
    m->popup(1, 0);
  }
}


void ControllerBox::showAboutDialog() {
  Dialog d("About MIDI Controller", true);
  Label l("<b>MIDI Controller " VERSION "</b>\n\n"
	  "This is a program that loads Glade UI "
	  "files and connects Range and SpinButton "
	  "widgets to MIDI Control Change values "
	  "that can be used to control synthesizers "
	  "and effects.\n\n"
	  "Report bugs, submit patches, and download "
	  "the latest version at\n\n"
	  "<b>http://www.sourceforge.net/midicontrol</b>\n\n"
	  "<i>(C) 2004 Lars Luthman "
	  "&lt;larsl@users.sourceforge.net&gt;</i>\n\n"
	  "MIDI Controller is free software and comes with "
	  "ABSOLUTELY NO WARRANTY; for details see the file COPYING.");
  l.set_use_markup(true);
  l.set_line_wrap(true);
  l.set_justify(JUSTIFY_CENTER);
  d.set_border_width(20);
  d.get_vbox()->pack_start(l);
  d.show_all();
  d.run();
}


void ControllerBox::savePreset() const {
  FileChooserDialog d(*win, "Save preset as...", FILE_CHOOSER_ACTION_SAVE);
  d.add_button(Stock::CANCEL, 0);
  d.add_button(Stock::OK, 1);
  d.set_default_response(1);
  if (d.run()) {
    cout<<"Saving preset to "<<d.get_filename()<<endl;
    map<string, Adjustment*>::const_iterator iter;
    ofstream preset(d.get_filename().c_str());
    for (iter = adjustments.begin(); iter != adjustments.end(); ++iter)
      preset<<iter->first<<'='<<iter->second->get_value()<<endl;
  }
}
  

void ControllerBox::loadPreset() {
  FileChooserDialog d(*win, "Load preset file");
  d.add_button(Stock::CANCEL, 0);
  d.add_button(Stock::OK, 1);
  d.set_default_response(1);
  if (d.run()) {
    
    // stop all controllers from sending CCs while we're changing them
    for (int i = 0; i < controllers.size(); ++i)
      controllers[i]->block();
    
    ifstream preset(d.get_filename().c_str());
    string str;
    while (getline(preset, str)) {
      size_t pos;
      if ((pos = str.find_last_of('=')) != -1) {
	map<string, Adjustment*>::iterator iter = 
	  adjustments.find(str.substr(0, pos));
	if (iter == adjustments.end())
	  cerr<<"Unknown widget: "<<iter->first<<endl;
	else {
	  istringstream iss(str.substr(pos + 1));
	  int d;
	  iss>>d;
	  Adjustment* adj = iter->second;
	  adj->set_value(d);
	  cout<<"Set "<<iter->first<<" to "<<d<<endl;
	}
      }
    }
    
    for (int i = 0; i < controllers.size(); ++i)
      controllers[i]->unblock();
  }
}


void ControllerBox::sendAll() const {
  for (int i = 0; i < controllers.size(); ++i)
    controllers[i]->sendControlChange();
}


Window* ControllerBox::getWindow() const {
  return win;
}



// CONTROLLERBOX::CONTROLLER IMPLEMENTATION

ControllerBox::Controller::Controller(Adjustment* valueAdj, 
				      snd_seq_t* sequencer)
  : valA(valueAdj), seq(sequencer) {
  conn = valueAdj->signal_value_changed().
    connect(sigc::mem_fun(this, &Controller::sendControlChange));
}


void ControllerBox::Controller::setChannel(unsigned char channel) {
  chn = channel;
}


void ControllerBox::Controller::setChannelAdj(Adjustment* channelAdj) {
  chn = (unsigned char)channelAdj->get_value();
  channelAdj->signal_value_changed().
    connect(sigc::mem_fun(this, &ControllerBox::Controller::onChannelChange));
  chanA = channelAdj;
}


void ControllerBox::Controller::setParameter(unsigned int parameter) { 
  par = parameter;
}


void ControllerBox::Controller::setParameterAdj(Adjustment* parameterAdj) {
  par = (unsigned int)parameterAdj->get_value();
  parameterAdj->signal_value_changed().
    connect(sigc::mem_fun(this, &ControllerBox::Controller::onParamChange));
  paramA = parameterAdj;
}


void ControllerBox::Controller::sendControlChange() const {
  snd_seq_event_t ev;
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_controller(&ev, chn - 1, par, (int)valA->get_value());
  snd_seq_ev_set_direct(&ev);
  int err = snd_seq_event_output_direct(seq, &ev);
  if (err < 0)
    cerr<<"Could not send MIDI control change: "<<snd_strerror(err)<<endl;
  else
    cerr<<"Sent "<<int(chn-1)<<':'<<par<<':'<<valA->get_value()<<endl;
}


void ControllerBox::Controller::block() {
  conn.block();
}


void ControllerBox::Controller::unblock() {
  conn.unblock();
  sendControlChange();
}


void ControllerBox::Controller::onChannelChange() {
  chn = (unsigned char)chanA->get_value();
}


void ControllerBox::Controller::onParamChange() {
  par = (unsigned int)paramA->get_value();
}



// GLOBAL FUNCTIONS

/** Print version and copyright info. */
void printVersion() {
  cout<<"MIDI Controller "<<VERSION<<endl
      <<"Copyright (C) 2004 Lars Luthman <larsl@users.sourceforge.net>"<<endl
      <<"This program comes with ABSOLUTELY NO WARRANTY."<<endl
      <<"This is free software, and you are welcome to redistribute it"<<endl
      <<"under certain conditions; see the file COPYING for details."<<endl;
}


int main(int argc, char** argv) {
  
  if (argc > 1 && !strcmp(argv[1], "--version")) {
    printVersion();
    return 0;
  }
  
  Main kit(argc, argv);
  try {
    ControllerBox box(argc > 1 ? argv[1] : "");
    Main::run(*(box.getWindow()));
  }
  catch (Glib::Exception& e) {
    cerr<<"Exception caught: "<<e.what()<<endl;
    return -1;
  }
  
  return 0;
}
